from typing import List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.vector import VectorResult

__all__ = ["QueryRequest", "VectorResponse"]


class QueryRequestBase(BaseModel):
    """
    A generic search/query against a collection.

    Accepts a SQL-compatible Metadata filter.
    Filter operations can be performed on any metadata column declared in the
    collection of supported data type.

    Supported data types: Boolean, Int, Long , Float, Double, String

    Supported filter operators (a subset of SQL:1999 standard)
    - Arithmetic Operators: + , - , * , / , %
    - Comparison Operators: =, <, >, <=, >=, !=
    - Boolean Operators: and, or, not
    - Grouping Operators: ()
    - Null Check: is null, is not null

    Note about filter syntax:
    - Do not enclose metadata column names
    - Do not enclose Int, Long, Float or Double values.
    - Enclose String values with single quote

    Examples:
    1. A Collection contains 'published_on (int)' metadata column that contains
       the epoch timestamp of when a vector is published. 'tag (string)' column
       that contains a tag for the vector. The following filter can be used to
       restricts query results to only important vectors published after
       1/1/2022 GMT.

       published_on > 1641024000 and tag = 'important'

    2. A Collection contains reddit comments with metadata columns 'karma (int)'
       . 'published_on (int)', 'char_count (int)'. The following could be a
       reasonable filter to restrict results to only recent, popular and
       informative comments.

       karma >= 50 and published_on > 1641024000  and char_count > 100

    3. Other examples of valid filters
       (foo_int > 10 or bar_string = 'front_page') and baz_string != 'draft'
       (a = 1 or (b = 2 and c = 3 and (d = 4 or e = 5)))
    """
    metadata_filter: Optional[str] = None

    # A list of metadata columns to return
    include_metadata: Optional[List[str]] = None

    # Whether embeddings are returned
    include_values: Optional[bool] = False

    # Number of results to return
    # Up to 1000 is supported. If a higher value is specified, results may be
    # truncated to 1000.
    top_k: Optional[int] = 3


class QueryRequest(QueryRequestBase):
    """
    A metadata query against a collection.
    """
    # See base class for documentation.
    metadata_filter: str = ""


class VectorResponse(BaseModel):
    """
    Result of a metadata or search query
    """
    vectors: List[VectorResult]



import os
from typing import List, Dict, Optional
from urllib import request

from pydantic import BaseModel

from nautilusdb.client_models.chat import Chat
from nautilusdb.client_models.column_type import ColumnType
from nautilusdb.client_models.app import AnswerReference
from nautilusdb.client_models.query import QueryRequest, VectorResponse
from nautilusdb.client_models.search import SearchRequest
from nautilusdb.server_models.app_api import (
    AddDocRequest,
    AskRequest,
    AskResponse,
    SummaryRequest,
    SummaryResponse,
)
from nautilusdb.server_models.collection_api import (
    DeleteVectorsRequest,
    DescribeCollectionResponse,
)
from nautilusdb.server_models.search_api import (
    SearchWithEmbedding,
    Query,
    SearchRequest as ServerSearchRequest,
    SearchResponse as ServerSearchResponse,
)
from nautilusdb.server_models.query_api import (
    QueryRequest as ServerQueryRequest,
    QueryResponse as ServerQueryResponse,
)
from nautilusdb.utils.ask import ask
from nautilusdb.utils.config import Config
from nautilusdb.client_models.vector import Vector
from nautilusdb.server_models.vector_api import (
    Vector as ServerVector,
    UpsertRequest,
    UpsertResponse,
)
from nautilusdb.utils.file_upload import (
    is_url,
    validate_file,
    validate_url_file,
)

__all__ = ["Collection"]


class Collection(BaseModel):
    """
    A Collection is a named vector search index with a fixed embedding dimension.
    """
    #  Name of the collection
    name: str

    # Dimension of the vectors in this collection
    dimension: int = 0

    # Description of the collection
    description: str = ""

    # A set of metadata columns associated with vectors in this collection.
    #       Key: Column name
    #       Value: Column type(avro primitive type)
    metadata_columns: Optional[Dict[str, ColumnType]] = None

    # Distance metric used for vector search.
    # Only 'l2' distance is supported.
    distance_metric: str = 'l2'

    def upsert_vector(self, vectors: List[Vector]) -> int:
        """
        Upserts a list of vectors.

        For every vector in the list:
        - If a vector with the same ID already exists in the collection,
          it will be replaced with the provided vector
        - If a vector with the same ID does not exist in the collection, the
          provided vector will be created.

        :param vectors: A list of vectors to create or update
        :return: The number of vectors created or updated
        """
        url = f'{Config.get_base_url()}/vectors/upsert'
        req = UpsertRequest(
            project_name=Config.project,
            collection_name=self.name,
            vectors=[ServerVector.from_client_vector(v) for v in vectors])
        resp = Config.post(url=url, data=req.model_dump_json())
        resp = UpsertResponse.model_validate(resp.json())
        return resp.upsert_count

    def delete_vectors(
        self,
        vector_ids: List[str] = None,
        metadata_filter: str = None,
        delete_all=False
    ):
        """
        Deletes vectors specified by the given list of vector IDs from the
        collection. Exactly one of `vector_ids`, `metadata_filter`, `delete_all`
        must be specified.

        `metadata_filter` is A SQL-compatible filter to specify vector deletion
        condition. For example, "foo_col > 1" indicates all vectors where the
        value of metadata column 'foo_col' greater than 1 should be deleted.

        All columns referenced in `metadata_filter` must exist in the
        collection, you can retrieve the list of columns via describe_collection
        API `ndb.describe_collection(<collection_name>)`.

        Simplified metadata filter syntax (EBNF grammar will be published soon):

        Supported filter operators (a subset of SQL:1999 standard)
        - Arithmetic Operators: + , - , * , / , %
        - Comparison Operators: =, <, >, <=, >=, !=
        - Boolean Operators: and, or, not
        - Grouping Operators: ()
        - Null Check: is null, is not null

        Note about filter syntax:
        - Do not enclose metadata column names
        - Do not enclose Int, Long, Float or Double values.
        - Enclose String values with single quote

        Examples:
        1. Deletes all vectors tagged with "draft" that were created before
           1/1/2022.

           created_on < 1641024000 and tag = 'draft'

        2. A Collection contains reddit comments with metadata columns
           'karma (int)', 'published_on (int)', 'char_count (int)'.
           The following could be a filter to remove all unpopular or very short
           comments.

           karma >= 50 or char_count > 10

        3. Other examples of valid filters
           (foo_int > 10 or bar_string = 'front_page') and baz_string != 'draft'
           (a = 1 or (b = 2 and c = 3 and (d = 4 or e = 5)))

        :param collection_name: name of the collection
        :param vector_ids: IDs of vectors to delete
        :param metadata_filter: A SQL-compatible filter to specify vector
                                deletion condition. For example, "foo_col > 1"
                                indicates all vectors where the value of
                                metadata column 'foo_col' greater than 1 should
                                be deleted.
        :param delete_all: Delete all vectors
        """
        specified_fields = (int(vector_ids is not None) +
                            int(metadata_filter is not None) +
                            int(delete_all))
        if specified_fields != 1:
            raise ValueError(
                "Exactly one of `vector_ids`, `metadata_filter` "
                "and `delete_all` must be specified")
        req = DeleteVectorsRequest(
            project_name=Config.project,
            collection_name=self.name,
            vector_ids=vector_ids,
            where=metadata_filter,
            delete_all=delete_all)
        url = f'{Config.get_base_url()}/vectors/delete'
        Config.post(url=url, data=req.model_dump_json())

    def upload_document(self, path_or_url: str):
        """
        Uploads a local or web document into this collection.

        :param path_or_url: Local file path or the URL of a document. URL must
                            contain scheme (http or https) prefix.
        :return:
        """
        url = f'{Config.get_base_url()}/qadocs/add'
        req = AddDocRequest(project_name=Config.project,
                            collection_name=self.name)
        if is_url(path_or_url):
            validate_url_file(path_or_url)
            with request.urlopen(path_or_url) as f:
                fname = os.path.basename(path_or_url)
                Config.post(
                    url=url, files={'file': (fname, f)},
                    data={'request': req.model_dump_json()})
        else:
            # Handle tilde "~/blah"
            path_or_url = os.path.expanduser(path_or_url)
            validate_file(path_or_url)
            with open(path_or_url, 'rb') as f:
                fname = os.path.basename(path_or_url)
                Config.post(
                    url=url, files={'file': (fname, f)},
                    data={'request': req.model_dump_json()})

    def summarize_document(self, doc_name) -> str:
        """
        Returns a summary of the document.

        :param path_or_url: Local path or url to the uploaded document
        :return: A summary of the document
        """
        url = f'{Config.get_base_url()}/qadocs/summary'
        req = SummaryRequest(
            project_name=Config.project,
            collection_name=self.name,
            doc_name=doc_name,
        )
        resp = Config.post(url=url, data=req.model_dump_json())
        resp = SummaryResponse.model_validate(resp.json())
        return resp.summary

    def ask(
        self,
        question: str,
    ) -> (str, List[AnswerReference]):
        """
        Ask a question within the context of this collection.

        :param question: Question in plain text.
        :return: A tuple
            - answer: A plain-text answer to the given question
            - references: A list of most relevant document references in the
                          collection
        """
        return ask(Config.project, self.name, question)

    def search(self, queries: List[SearchRequest]) -> List[VectorResponse]:
        """
        Searches the collection
        """
        req = ServerSearchRequest(
            project_name=Config.project,
            collection_name=self.name,
            queries=[
                SearchWithEmbedding.from_client_request(q) for q in queries])

        url = f'{Config.get_base_url()}/vectors/search'

        resp = Config.post(url=url, data=req.model_dump_json())
        resp = ServerSearchResponse.model_validate(resp.json())
        return [r.to_client_response() for r in resp.results]

    def query(self, queries: List[QueryRequest]) -> List[VectorResponse]:
        """
        Queries the collection
        """
        req = ServerQueryRequest(
            project_name=Config.project,
            collection_name=self.name,
            queries=[
                Query.from_client_request(q) for q in queries])

        url = f'{Config.get_base_url()}/vectors/query'

        resp = Config.post(url=url, data=req.model_dump_json())
        resp = ServerQueryResponse.model_validate(resp.json())
        return [r.to_client_response() for r in resp.results]

    def chat(self, doc_name=None) -> Chat:
        """
        Creates a new chatbot within the collection.

        :param doc_name: If specified, chatbot is scoped to the given document only
        :return: A new Chat object
        """
        return Chat(
            project_name=Config.project,
            collection_name=self.name,
            doc_name=doc_name
        )

def describe_response_to_collection(
    resp: DescribeCollectionResponse
) -> Collection:
    return Collection(
        name=resp.collection_name,
        dimension=resp.dimension,
        metadata_columns=resp.metas,
    )



from typing import List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.query import QueryRequestBase


__all__ = ["SearchRequest"]


class SearchRequest(QueryRequestBase):
    """
    A search against a collection.
    """
    # Embedding of the query
    embedding: List[float]



from typing import List

from pydantic import BaseModel

from nautilusdb.server_models.collection_api import CollectionBaseRequest
from nautilusdb.server_models.search_api import VectorResult, Query


class QueryRequest(CollectionBaseRequest):
    queries: List[Query]


class QueryResponse(BaseModel):
    # Return a list of search results.
    # In the future, query may include similarity scores
    results: List[VectorResult]


from typing import List

from nautilusdb.server_models.app_api import AskRequest, AskResponse, AnswerReference
from nautilusdb.utils.config import Config


def ask(
    project_name: str,
    collection_name: str,
    question: str,
    doc_name=None,
    history_messages=None,
) -> (str, List[AnswerReference]):
    url = f'{Config.get_base_url()}/qadocs/ask'

    req = AskRequest(
        project_name=project_name,
        collection_name=collection_name,
        doc_name=doc_name,
        question=question,
        history_messages=history_messages,
    )
    enforce_api_key = True

    # Only allow openai-web collection to be accessed without API key to
    # simplify demo.
    if collection_name == 'openai-web':
        enforce_api_key = False

    resp = Config.post(
        url=url,
        data=req.model_dump_json(),
        enforce_api_key=enforce_api_key
    )
    resp = AskResponse.model_validate(resp.json())

    return (
        resp.answer,
        [AnswerReference(doc_name=ref.doc_name,
                         page_num=ref.page_num,
                         sample_text=ref.sample_text) for ref in resp.refs]
    )



from typing import Optional, List, Tuple

from pydantic import BaseModel

from nautilusdb.client_models.app import AnswerReference
from nautilusdb.utils.ask import ask
from nautilusdb.utils.config import Config


class Chat(BaseModel):
    """
    A new chatbot instance that supports conversational Q and A.
    """
    project_name: str
    collection_name: str
    doc_name: Optional[str] = None
    history_messages: Optional[List[Tuple[str, str]]] = None

    def ask(self, question: str) -> (str, List[AnswerReference]):
        (answer, references) = ask(
            Config.project,
            self.collection_name,
            question,
            history_messages=self.history_messages
        )
        if self.history_messages is None:
            self.history_messages = []
        self.history_messages.append((question, answer))
        return answer, references



import json
from typing import List, Optional, Tuple

from pydantic import BaseModel

from nautilusdb.server_models.collection_api import CollectionBaseRequest


class AddDocRequest(CollectionBaseRequest):
    @classmethod
    def validate_to_json(cls, value):
        return cls(**json.loads(value)) if isinstance(value, str) else value

    def to_json_string(self) -> str:
        # fastapi server expects "property name enclosed in double quotes" when
        # using with UploadFile. pydantic.model_dump_json() uses single quote.
        # explicitly uses json.dumps for AddDocRequest.
        return json.dumps(self.__dict__)


class AskRequest(CollectionBaseRequest):
    question: str
    top_k: Optional[int] = 5
    doc_name: Optional[str] = None
    history_messages: Optional[List[Tuple[str, str]]] = None


class AnswerReference(BaseModel):
    doc_name: str
    page_num: int
    sample_text: str


class AskResponse(BaseModel):
    answer: str
    refs: List[AnswerReference]


class SummaryRequest(CollectionBaseRequest):
    doc_name: str


class SummaryResponse(BaseModel):
    summary: str



from typing import List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.search import (
    SearchRequest as ClientSearchRequest,
)
from nautilusdb.client_models.query import (
    QueryRequest as ClientQueryRequest,
    VectorResponse as ClientVectorResponse,
)
from nautilusdb.server_models.collection_api import CollectionBaseRequest
from nautilusdb.server_models.vector_api import VectorWithScore


class VectorResult(BaseModel):
    vectors: List[VectorWithScore]

    def to_client_response(self) -> ClientVectorResponse:
        return ClientVectorResponse(
            vectors=[v.to_client_vector_request() for v in self.vectors])


class Query(BaseModel):
    # where supports SQL =, <, >, <=, >=, !=, and, or, etc.
    where: Optional[str] = None
    return_metas: Optional[List[str]] = None
    return_embedding: Optional[bool] = False
    top_k: Optional[int] = 3

    @classmethod
    def from_client_request(cls, req: ClientQueryRequest) -> "Query":
        return Query(
            where=req.metadata_filter,
            return_metas=req.include_metadata,
            return_embedding=req.include_values,
            top_k=req.top_k
        )


class SearchWithEmbedding(Query):
    embedding: List[float]

    @classmethod
    def from_client_request(
        cls,
        req: ClientSearchRequest
    ) -> "SearchWithEmbedding":
        return SearchWithEmbedding(
            embedding=req.embedding,
            where=req.metadata_filter,
            return_metas=req.include_metadata,
            return_embedding=req.include_values,
            top_k=req.top_k
        )


class SearchRequest(CollectionBaseRequest):
    queries: List[SearchWithEmbedding]


class SearchResponse(BaseModel):
    # The results are sorted by the score in the descending order
    results: List[VectorResult]



from typing import Dict, List, Optional

from pydantic import BaseModel

from nautilusdb.client_models.column_type import ColumnType


class CollectionBaseRequest(BaseModel):
    project_name: str
    collection_name: str


class CreateCollectionRequest(CollectionBaseRequest):
    dimension: int
    description: Optional[str] = None
    metas: Optional[Dict[str, ColumnType]] = None


class DeleteCollectionRequest(CollectionBaseRequest):
    pass


class ListCollectionsRequest(BaseModel):
    project_name: str


class ListCollectionsResponse(BaseModel):
    names: List[str]


class DescribeCollectionRequest(CollectionBaseRequest):
    pass


class DescribeCollectionResponse(BaseModel):
    collection_name: str
    dimension: int
    metric: str = 'l2'
    metas: Optional[Dict[str, ColumnType]] = None


class DeleteVectorsRequest(CollectionBaseRequest):
    """
    Delete vectors in a collection by ids, all or by where clause. Exactly one
    of these three fields can be specified.
    """
    vector_ids: Optional[List[str]] = None
    delete_all: bool = False
    where: Optional[str] = None


class DeleteVectorsResponse(BaseModel):
    pass



import requests
import os

from .exceptions import raise_for_error

ENV_TEST_BASE_URL = os.environ.get("ENV_TEST_BASE_URL", "")

class Config:
    PUBLIC_ACCOUNT: str = "public"
    PUBLIC_PROJECT: str = "public"
    api_key: str = None
    project: str = PUBLIC_PROJECT
    account_name: str = PUBLIC_ACCOUNT

    @classmethod
    def validate_setup(cls):
        if cls.account_name is None or cls.account_name != cls.PUBLIC_ACCOUNT:
            raise ValueError("Only public account ('public') is supported.")

    @classmethod
    def get_base_url(cls):
        if ENV_TEST_BASE_URL != "":
            return ENV_TEST_BASE_URL
        return f"https://{cls.account_name}.us-west-2.aws.nautilusdb.com"

    @classmethod
    def post(cls, url, data=None, files=None, enforce_api_key=True):
        if enforce_api_key and cls.api_key is None or cls.api_key == "":
            raise ValueError("API key is required for using NautilusDB")

        # Encode data if it is a string
        if data is not None and isinstance(data, str):
            data = data.encode('utf-8')
        resp = requests.post(
            url, data=data, files=files, headers=cls._construct_header())
        raise_for_error(resp)
        return resp

    @classmethod
    def get(cls, url: str):
        if cls.api_key is None or cls.api_key == "":
            raise ValueError("API key is required for using NautilusDB")

        resp = requests.get(url, headers=cls._construct_header())
        raise_for_error(resp)
        return resp

    @classmethod
    def _construct_header(cls):
        return None if cls.api_key is None else {'api-key': cls.api_key}
